from typing import (
    Callable,
    Dict,
    Optional,
    Tuple,
    Type
)

from .registry import (
    Meta,
    Registry
)
from .scope import Scope
from .type import B, T, Tag


class Container:

    @classmethod
    def create(cls, parent: Optional["Container"] = None) -> "Container":
        return cls(parent=parent)

    def __init__(self, parent: Optional["Container"] = None):
        self._parent = parent
        self._registry = Registry()

    # public

    def create_scope(self) -> "Container":
        return Container.create(parent=self)

    def get(self, target: Type[T], *, tag: Optional[Tag] = None) -> T:
        meta, is_self = self._get_meta(target, tag=tag)

        # return cache instance when scope is Singleton
        if meta.scope == Scope.Singleton and meta.instance:
            return meta.instance

        # return cache instance in current container when scope is Scope
        if meta.scope == Scope.Scope and meta.instance and is_self:
            return meta.instance

        instance = self._create_instance(meta)

        # cache instance when scope is Singleton
        if meta.scope == Scope.Singleton:
            meta.instance = instance

        # cache instance in current container when scope is Scope
        if meta.scope == Scope.Scope:
            scope_meta = meta.copy()
            scope_meta.instance = instance
            self._registry.register_meta(
                scope_meta,
                on_behalf_of=target,
                tag=tag
            )

        if instance:
            return instance

        # generic type T is missing when function returns,
        # instance must exist, or it will raise exception while creating instance,
        # this line is only for patching missing generic type T for this function.
        return target()

    def register(
        self,
        target: Type[T],
        *,
        factory: Optional[Callable[[], T]] = None,
        on_behalf_of: Optional[Type[B]] = None,
        scope: Optional[Scope] = Scope.Transient,
        tag: Optional[Tag] = None,
        value: Optional[T] = None,
        params: Optional[Dict[str, str]] = None
    ) -> None:
        self._registry.register(
            target,
            on_behalf_of=on_behalf_of,
            factory=factory,
            scope=scope,
            tag=tag,
            value=value,
            params=params,
        )

    def reset(self):
        self._registry = Registry()

    # private

    def _create_instance(self, meta: Meta[T]) -> T:
        if meta.factory:
            return meta.factory()

        if meta.value:
            return meta.value

        args = []
        for _, p in meta.signature.parameters.items():
            arg = self.get(p.annotation, tag=meta.get_tag(p))
            args.append(arg)

        return meta.ctor(*args)

    def _get_meta(self, target: Type[T], *, tag: Optional[Tag] = None) -> Tuple[Meta[T], bool]:
        try:
            is_self = True
            return self._registry.get(target, tag=tag), is_self
        except KeyError as e:
            if not self._parent:
                raise e
            is_self = False
            meta, _ = self._parent._get_meta(target, tag=tag)
            return meta, is_self


container = Container.create()



from typing import Any, Type


def is_builtin_type(target: Type[Any]) -> bool:
    try:
        # https://stackoverflow.com/a/17795199
        return target.__module__ in [
            "__builtin__",
            "__builtins__",
            "builtins"
        ]
    except KeyError as e:
        return False



from typing import (
    Callable,
    Dict,
    Optional,
    Type
)

from .di import (
    Container,
    container as c
)
from .scope import Scope
from .type import B, T, Tag


def injectable(
    on_behalf_of: Optional[Type[B]] = None,
    *,
    container: Optional[Container] = None,
    scope: Optional[Scope] = None,
    tag: Optional[Tag] = None,
    params: Optional[Dict[str, str]] = None
) -> Callable[[Type[T]], Type[T]]:
    def wrap(target: Type[T]) -> Type[T]:
        (container or c).register(
            target,
            on_behalf_of=on_behalf_of,
            scope=scope,
            tag=tag,
            params=params
        )
        return target
    return wrap



import inspect

from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Type
)

from .scope import Scope
from .type import B, T, Tag
from .util import is_builtin_type


class Meta(Generic[T]):

    def __init__(
        self,
        ctor: Type[T],
        *,
        scope: Optional[Scope] = Scope.Transient,
        factory: Optional[Callable[[], T]] = None,
        value: Optional[T] = None,
        params: Optional[Dict[str, str]] = None
    ):
        self.ctor = ctor
        self.scope = scope
        self.factory = factory
        self.value = value
        self.params = params or {}

        # derived properies

        self.instance: Optional[T] = None

        if is_builtin_type(ctor):
            self.signature = inspect.Signature()
        else:
            self.signature = inspect.signature(ctor)

    def copy(self) -> "Meta[T]":
        meta = Meta[T](
            self.ctor,
            scope=self.scope,
            factory=self.factory,
            value=self.value,
            params={**self.params}
        )
        return meta

    def get_tag(self, param: inspect.Parameter) -> Optional[Tag]:
        if param.name in self.params:
            return self.params[param.name]

        if is_builtin_type(param.annotation):
            return param.name

        return None


class Registry:

    def __init__(self):
        self._registry: Dict[Type[Any], Dict[Tag, Meta[Any]]] = {}

    # public

    def get(
        self,
        target: Type[T],
        *,
        tag: Optional[Tag] = None
    ) -> Meta[T]:
        return self._registry[target][self._get_impl_key(tag)]

    def register(
        self,
        target: Type[T],
        *,
        factory: Optional[Callable[[], T]] = None,
        on_behalf_of: Optional[Type[B]] = None,
        scope: Optional[Scope] = Scope.Transient,
        tag: Optional[Tag] = None,
        value: Optional[T] = None,
        params: Optional[Dict[str, str]] = None
    ) -> None:
        target_key = on_behalf_of or target
        if target_key not in self._registry:
            self._registry[target_key] = {}
        self._registry[target_key][self._get_impl_key(tag)] = Meta(
            target,
            scope=scope,
            factory=factory,
            value=value,
            params=params
        )

    def register_meta(
        self,
        meta: Meta[Any],
        *,
        on_behalf_of: Optional[Type[B]] = None,
        tag: Optional[Tag] = None
    ) -> None:
        target_key = on_behalf_of or meta.ctor
        if target_key not in self._registry:
            self._registry[target_key] = {}
        self._registry[target_key][self._get_impl_key(tag)] = meta

    # private

    def _get_impl_key(self, tag: Optional[Tag] = None) -> Tag:
        return tag if tag else "DEFAULT"
